@@ -30,7 +30,7 @@
 #include "zend_dump.h"
 
 /* Checks if a constant (like "true") may be replaced by its value */
-int zend_optimizer_get_persistent_constant(zend_string *name, zval *result, int copy)
+bool zend_optimizer_get_persistent_constant(zend_string *name, zval *result, int copy)
 {
 	zend_constant *c = zend_hash_find_ptr(EG(zend_constants), name);
 	if (c) {
@@ -174,11 +174,8 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 					 && opline->opcode != ZEND_MATCH
 					 && zend_optimizer_update_op1_const(op_array, opline, &c)) {
 						VAR_SOURCE(op1) = NULL;
-						if (opline->opcode != ZEND_JMP_NULL
-						 && !zend_bitset_in(used_ext, VAR_NUM(src->result.var))) {
-							literal_dtor(&ZEND_OP1_LITERAL(src));
-							MAKE_NOP(src);
-						}
+						literal_dtor(&ZEND_OP1_LITERAL(src));
+						MAKE_NOP(src);
 						++(*opt_count);
 					} else {
 						zval_ptr_dtor_nogc(&c);
@@ -200,10 +197,8 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 				ZVAL_COPY(&c, &ZEND_OP1_LITERAL(src));
 				if (zend_optimizer_update_op2_const(op_array, opline, &c)) {
 					VAR_SOURCE(op2) = NULL;
-					if (!zend_bitset_in(used_ext, VAR_NUM(src->result.var))) {
-						literal_dtor(&ZEND_OP1_LITERAL(src));
-						MAKE_NOP(src);
-					}
+					literal_dtor(&ZEND_OP1_LITERAL(src));
+					MAKE_NOP(src);
 					++(*opt_count);
 				} else {
 					zval_ptr_dtor_nogc(&c);
@@ -231,7 +226,7 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 						 * Float to string conversion may be affected by current
 						 * locale setting.
 						 */
-						int l, old_len;
+						size_t l, old_len;
 
 						if (Z_TYPE(ZEND_OP1_LITERAL(opline)) != IS_STRING) {
 							convert_to_string(&ZEND_OP1_LITERAL(opline));
@@ -261,11 +256,14 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 				}
 				break;
 
+			case ZEND_MATCH_ERROR:
+				if (opline->op1_type & (IS_TMP_VAR|IS_VAR)) {
+					src = VAR_SOURCE(opline->op1);
+					VAR_SOURCE(opline->op1) = NULL;
+				}
+				break;
+
 			case ZEND_FREE:
-				/* Note: Only remove the source if the source is local to this block.
-				 * If it's not local, then the other blocks successors must also eventually either FREE or consume the temporary,
-				 * hence removing the temporary is not safe in the general case, especially when other consumers are not FREE.
-				 * A FREE may not be removed without also removing the source's result, because otherwise that would cause a memory leak. */
 				if (opline->op1_type == IS_TMP_VAR) {
 					src = VAR_SOURCE(opline->op1);
 					if (src) {
@@ -274,7 +272,6 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 							case ZEND_BOOL_NOT:
 								/* T = BOOL(X), FREE(T) => T = BOOL(X) */
 								/* The remaining BOOL is removed by a separate optimization */
-								/* The source is a bool, no source removals take place, so this may be done non-locally. */
 								VAR_SOURCE(opline->op1) = NULL;
 								MAKE_NOP(opline);
 								++(*opt_count);
@@ -293,9 +290,6 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 							case ZEND_PRE_DEC_OBJ:
 							case ZEND_PRE_INC_STATIC_PROP:
 							case ZEND_PRE_DEC_STATIC_PROP:
-								if (src < op_array->opcodes + block->start) {
-									break;
-								}
 								src->result_type = IS_UNUSED;
 								VAR_SOURCE(opline->op1) = NULL;
 								MAKE_NOP(opline);
@@ -308,7 +302,7 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 				} else if (opline->op1_type == IS_VAR) {
 					src = VAR_SOURCE(opline->op1);
 					/* V = OP, FREE(V) => OP. NOP */
-					if (src >= op_array->opcodes + block->start &&
+					if (src &&
 					    src->opcode != ZEND_FETCH_R &&
 					    src->opcode != ZEND_FETCH_STATIC_PROP_R &&
 					    src->opcode != ZEND_FETCH_DIM_R &&
@@ -347,7 +341,7 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 			   ) {
 				zval *arg = &OPLINE_OP1_LITERAL(sv);
 				char *fname = FUNCTION_CACHE->funcs[Z_LVAL(ZEND_OP1_LITERAL(fcall))].function_name;
-				int flen = FUNCTION_CACHE->funcs[Z_LVAL(ZEND_OP1_LITERAL(fcall))].name_len;
+				size_t flen = FUNCTION_CACHE->funcs[Z_LVAL(ZEND_OP1_LITERAL(fcall))].name_len;
 				if((flen == sizeof("function_exists")-1 && zend_binary_strcasecmp(fname, flen, "function_exists", sizeof("function_exists")-1) == 0) ||
 						  (flen == sizeof("is_callable")-1 && zend_binary_strcasecmp(fname, flen, "is_callable", sizeof("is_callable")-1) == 0)
 						  ) {
@@ -361,7 +355,7 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 					}
 				} else if(flen == sizeof("constant")-1 && zend_binary_strcasecmp(fname, flen, "constant", sizeof("constant")-1) == 0) {
 					zval c;
-					if(zend_optimizer_get_persistent_constant(Z_STR_P(arg), &c, 1 ELS_CC) != 0) {
+					if (zend_optimizer_get_persistent_constant(Z_STR_P(arg), &c, 1 ELS_CC)) {
 						literal_dtor(arg);
 						MAKE_NOP(sv);
 						MAKE_NOP(fcall);
@@ -633,53 +627,6 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 				}
 				break;
 
-			case ZEND_JMPZNZ:
-				while (1) {
-					if (opline->op1_type == IS_CONST) {
-						++(*opt_count);
-						if (zend_is_true(&ZEND_OP1_LITERAL(opline))) {
-							zend_op *target_opline = ZEND_OFFSET_TO_OPLINE(opline, opline->extended_value);
-							ZEND_SET_OP_JMP_ADDR(opline, opline->op1, target_opline);
-							block->successors[0] = block->successors[1];
-						} else {
-							zend_op *target_opline = ZEND_OP2_JMP_ADDR(opline);
-							ZEND_SET_OP_JMP_ADDR(opline, opline->op1, target_opline);
-						}
-						block->successors_count = 1;
-						opline->op1_type = IS_UNUSED;
-						opline->extended_value = 0;
-						opline->opcode = ZEND_JMP;
-						break;
-					} else if (opline->op1_type == IS_TMP_VAR &&
-					           !zend_bitset_in(used_ext, VAR_NUM(opline->op1.var))) {
-						src = VAR_SOURCE(opline->op1);
-						if (src) {
-							if (src->opcode == ZEND_BOOL_NOT) {
-								/* T = BOOL_NOT(X) + JMPZNZ(T,L1,L2) -> NOP, JMPZNZ(X,L2,L1) */
-								uint32_t tmp;
-
-								VAR_SOURCE(opline->op1) = NULL;
-								COPY_NODE(opline->op1, src->op1);
-								tmp = block->successors[0];
-								block->successors[0] = block->successors[1];
-								block->successors[1] = tmp;
-								MAKE_NOP(src);
-								++(*opt_count);
-								continue;
-							} else if (src->opcode == ZEND_BOOL ||
-							           src->opcode == ZEND_QM_ASSIGN) {
-								VAR_SOURCE(opline->op1) = NULL;
-								COPY_NODE(opline->op1, src->op1);
-								MAKE_NOP(src);
-								++(*opt_count);
-								continue;
-							}
-						}
-					}
-					break;
-				}
-				break;
-
 			case ZEND_JMPZ_EX:
 			case ZEND_JMPNZ_EX:
 				while (1) {
@@ -732,7 +679,7 @@ static void zend_optimize_block(zend_basic_block *block, zend_op_array *op_array
 					     src->opcode == ZEND_FAST_CONCAT) &&
 					    src->op2_type == IS_CONST) {
 						/* compress consecutive CONCATs */
-						int l, old_len;
+						size_t l, old_len;
 
 						if (Z_TYPE(ZEND_OP2_LITERAL(opline)) != IS_STRING) {
 							convert_to_string(&ZEND_OP2_LITERAL(opline));
@@ -980,7 +927,6 @@ static void assemble_code_blocks(zend_cfg *cfg, zend_op_array *op_array, zend_op
 	zend_op *new_opcodes;
 	zend_op *opline;
 	uint32_t len = 0;
-	int n;
 
 	for (b = blocks; b < end; b++) {
 		if (b->len == 0) {
@@ -1046,9 +992,6 @@ static void assemble_code_blocks(zend_cfg *cfg, zend_op_array *op_array, zend_op
 			case ZEND_JMP:
 				ZEND_SET_OP_JMP_ADDR(opline, opline->op1, new_opcodes + blocks[b->successors[0]].start);
 				break;
-			case ZEND_JMPZNZ:
-				opline->extended_value = ZEND_OPLINE_TO_OFFSET(opline, new_opcodes + blocks[b->successors[1]].start);
-				ZEND_FALLTHROUGH;
 			case ZEND_JMPZ:
 			case ZEND_JMPNZ:
 			case ZEND_JMPZ_EX:
@@ -1140,16 +1083,9 @@ static void assemble_code_blocks(zend_cfg *cfg, zend_op_array *op_array, zend_op
 		free_alloca(map, use_heap);
 	}
 
-	/* adjust early binding list */
-	if (op_array->fn_flags & ZEND_ACC_EARLY_BINDING) {
-		ZEND_ASSERT(op_array == &ctx->script->main_op_array);
-		ctx->script->first_early_binding_opline =
-			zend_build_delayed_early_binding_list(op_array);
-	}
-
 	/* rebuild map (just for printing) */
 	memset(cfg->map, -1, sizeof(int) * op_array->last);
-	for (n = 0; n < cfg->blocks_count; n++) {
+	for (int n = 0; n < cfg->blocks_count; n++) {
 		if (cfg->blocks[n].flags & (ZEND_BB_REACHABLE|ZEND_BB_UNREACHABLE_FREE)) {
 			cfg->map[cfg->blocks[n].start] = n;
 		}
@@ -1209,7 +1145,7 @@ static zend_always_inline zend_basic_block *get_next_block(const zend_cfg *cfg,
 
 
 /* we use "jmp_hitlist" to avoid infinity loops during jmp optimization */
-static zend_always_inline int in_hitlist(int target, int *jmp_hitlist, int jmp_hitlist_count)
+static zend_always_inline bool in_hitlist(int target, int *jmp_hitlist, int jmp_hitlist_count)
 {
 	int i;
 
@@ -1267,20 +1203,7 @@ static void zend_jmp_optimization(zend_basic_block *block, zend_op_array *op_arr
 				block->len--;
 			} else if (target_block->len == 1) {
 				target = op_array->opcodes + target_block->start;
-				if (target->opcode == ZEND_JMPZNZ) {
-					/* JMP L, L: JMPZNZ L1,L2 -> JMPZNZ L1,L2 */
-					*last_op = *target;
-					if (last_op->op1_type == IS_CONST) {
-						zval zv;
-						ZVAL_COPY(&zv, &ZEND_OP1_LITERAL(last_op));
-						last_op->op1.constant = zend_optimizer_add_literal(op_array, &zv);
-					}
-					block->successors_count = 2;
-					block->successors[0] = target_block->successors[0];
-					block->successors[1] = target_block->successors[1];
-					++(*opt_count);
-					goto optimize_jmpznz;
-				} else if ((target->opcode == ZEND_RETURN ||
+				if ((target->opcode == ZEND_RETURN ||
 				            target->opcode == ZEND_RETURN_BY_REF ||
 				            target->opcode == ZEND_GENERATOR_RETURN ||
 				            target->opcode == ZEND_EXIT) &&
@@ -1339,10 +1262,6 @@ static void zend_jmp_optimization(zend_basic_block *block, zend_op_array *op_arr
 				           SAME_VAR(target->op1, last_op->op1)) {
 					/* JMPZ(X, L), L: JMPNZ(X, L2) -> JMPZ(X, L+1) */
 					next = target_block->successors[1];
-				} else if (target->opcode == ZEND_JMPZNZ &&
-				           SAME_VAR(target->op1, last_op->op1)) {
-					/* JMPZ(X, L), L: JMPZNZ(X, L2, L3) -> JMPZ(X, L2) */
-					next = target_block->successors[last_op->opcode == ZEND_JMPNZ];
 				} else {
 					break;
 				}
@@ -1355,14 +1274,8 @@ static void zend_jmp_optimization(zend_basic_block *block, zend_op_array *op_arr
 			follow_block = get_follow_block(cfg, block, 1, opt_count);
 			if (target_block == follow_block) {
 				/* L: JMP[N]Z(X, L+1) -> NOP or FREE(X) */
-				if (last_op->op1_type == IS_CV) {
-					last_op->opcode = ZEND_CHECK_VAR;
-					last_op->op2.num = 0;
-				} else if (last_op->op1_type & (IS_VAR|IS_TMP_VAR)) {
-					last_op->opcode = ZEND_FREE;
-					last_op->op2.num = 0;
-				} else {
-					MAKE_NOP(last_op);
+				zend_optimizer_convert_to_free_op1(op_array, last_op);
+				if (last_op->opcode == ZEND_NOP) {
 					block->len--;
 				}
 				block->successors_count = 1;
@@ -1372,14 +1285,8 @@ static void zend_jmp_optimization(zend_basic_block *block, zend_op_array *op_arr
 				if (target->opcode == ZEND_JMP) {
 				    if (block->successors[0] == follow_block->successors[0]) {
 						/* JMPZ(X,L1), JMP(L1) -> NOP, JMP(L1) */
-						if (last_op->op1_type == IS_CV) {
-							last_op->opcode = ZEND_CHECK_VAR;
-							last_op->op2.num = 0;
-						} else if (last_op->op1_type & (IS_VAR|IS_TMP_VAR)) {
-							last_op->opcode = ZEND_FREE;
-							last_op->op2.num = 0;
-						} else {
-							MAKE_NOP(last_op);
+						zend_optimizer_convert_to_free_op1(op_array, last_op);
+						if (last_op->opcode == ZEND_NOP) {
 							block->len--;
 						}
 						block->successors[0] = follow_block - cfg->blocks;
@@ -1406,16 +1313,6 @@ static void zend_jmp_optimization(zend_basic_block *block, zend_op_array *op_arr
 							break;
 						}
 					}
-
-					/* JMPZ(X,L1), JMP(L2) -> JMPZNZ(X,L1,L2) */
-					if (last_op->opcode == ZEND_JMPZ) {
-						block->successors[1] = follow_block->successors[0];
-					} else {
-						block->successors[1] = block->successors[0];
-						block->successors[0] = follow_block->successors[0];
-					}
-					last_op->opcode = ZEND_JMPZNZ;
-					++(*opt_count);
 				}
 			}
 			break;
@@ -1442,11 +1339,6 @@ static void zend_jmp_optimization(zend_basic_block *block, zend_op_array *op_arr
 				            SAME_VAR(target->op1, last_op->op1))) {
 					/* T = JMPZ_EX(X, L1), L1: T = JMPZ_EX({X|T}, L2) -> T = JMPZ_EX(X, L2) */
 					next = target_block->successors[0];
-				} else if (target->opcode == ZEND_JMPZNZ &&
-				           (SAME_VAR(target->op1, last_op->result) ||
-				            SAME_VAR(target->op1, last_op->op1))) {
-					/* T = JMPZ_EX(X, L), L: JMPZNZ({X|T}, L2, L3) -> T = JMPZ_EX(X, L2) */
-					next = target_block->successors[last_op->opcode == ZEND_JMPNZ_EX];
 				} else if (target->opcode == INV_EX_COND(last_op->opcode) &&
 				           (SAME_VAR(target->op1, last_op->result) ||
 				            SAME_VAR(target->op1, last_op->op1))) {
@@ -1493,85 +1385,6 @@ static void zend_jmp_optimization(zend_basic_block *block, zend_op_array *op_arr
 				break;
 			}
 			break;
-
-		case ZEND_JMPZNZ: {
-optimize_jmpznz:
-			jmp_hitlist_count = 0;
-			target_block = get_target_block(cfg, block, 0, opt_count);
-			while (target_block->len == 1) {
-				target = op_array->opcodes + target_block->start;
-
-				if (target->opcode == ZEND_JMP) {
-					/* JMPZNZ(X, L1, L2), L1: JMP(L3) -> JMPZNZ(X, L3, L2) */
-					next = target_block->successors[0];
-				} else if ((target->opcode == ZEND_JMPZ || target->opcode == ZEND_JMPZNZ) &&
-				           SAME_VAR(target->op1, last_op->op1)) {
-					/* JMPZNZ(X, L1, L2), L1: JMPZ(X, L3) -> JMPZNZ(X, L3, L2) */
-					next = target_block->successors[0];
-				} else if (target->opcode == ZEND_JMPNZ &&
-				           SAME_VAR(target->op1, last_op->op1)) {
-					/* JMPZNZ(X, L1, L2), L1: X = JMPNZ(X, L3) -> JMPZNZ(X, L1+1, L2) */
-					next = target_block->successors[1];
-				} else {
-					break;
-				}
-				CHECK_LOOP(next);
-				block->successors[0] = next;
-				++(*opt_count);
-				target_block = get_target_block(cfg, block, 0, opt_count);
-			}
-
-			jmp_hitlist_count = 0;
-			follow_block = get_target_block(cfg, block, 1, opt_count);
-			while (follow_block->len == 1) {
-				target = op_array->opcodes + follow_block->start;
-
-				if (target->opcode == ZEND_JMP) {
-					/* JMPZNZ(X, L1, L2), L2: JMP(L3) -> JMPZNZ(X, L1, L3) */
-					next = follow_block->successors[0];
-				} else if (target->opcode == ZEND_JMPNZ &&
-				           SAME_VAR(target->op1, last_op->op1)) {
-					/* JMPZNZ(X, L1, L2), L2: X = JMPNZ(X, L3) -> JMPZNZ(X, L1, L3) */
-					next = follow_block->successors[0];
-				} else if ((target->opcode == ZEND_JMPZ || target->opcode == ZEND_JMPZNZ) &&
-				           SAME_VAR(target->op1, last_op->op1)) {
-					/* JMPZNZ(X, L1, L2), L2: JMPZ(X, L3) -> JMPZNZ(X, L1, L2+1) */
-					next = follow_block->successors[1];
-				} else {
-					break;
-				}
-				CHECK_LOOP(next);
-				block->successors[1] = next;
-				++(*opt_count);
-				follow_block = get_target_block(cfg, block, 1, opt_count);
-			}
-
-			next_block = get_next_block(cfg, block);
-			if (target_block == follow_block &&
-			    !(last_op->op1_type & (IS_VAR|IS_TMP_VAR))) {
-				/* JMPZNZ(?,L,L) -> JMP(L) */
-				last_op->opcode = ZEND_JMP;
-				SET_UNUSED(last_op->op1);
-				SET_UNUSED(last_op->op2);
-				last_op->extended_value = 0;
-				block->successors_count = 1;
-				++(*opt_count);
-			} else if (target_block == next_block) {
-				/* jumping to next on Z - can follow to it and jump only on NZ */
-				/* JMPZNZ(X,L1,L2) L1: -> JMPNZ(X,L2) */
-				int tmp = block->successors[0];
-				last_op->opcode = ZEND_JMPNZ;
-				block->successors[0] = block->successors[1];
-				block->successors[1] = tmp;
-				++(*opt_count);
-			} else if (follow_block == next_block) {
-				/* jumping to next on NZ - can follow to it and jump only on Z */
-				/* JMPZNZ(X,L1,L2) L2: -> JMPZ(X,L1) */
-				last_op->opcode = ZEND_JMPZ;
-				++(*opt_count);
-			}
-			break;
-		}
 	}
 }
 
@@ -1662,7 +1475,7 @@ static void zend_t_usage(zend_cfg *cfg, zend_op_array *op_array, zend_bitset use
 	}
 
 	if (ctx->debug_level & ZEND_DUMP_BLOCK_PASS_VARS) {
-		int printed = 0;
+		bool printed = 0;
 		uint32_t i;
 
 		for (i = op_array->last_var; i< op_array->T; i++) {
@@ -1732,18 +1545,7 @@ static void zend_t_usage(zend_cfg *cfg, zend_op_array *op_array, zend_bitset use
 						case ZEND_QM_ASSIGN:
 						case ZEND_BOOL:
 						case ZEND_BOOL_NOT:
-							if (opline->op1_type == IS_CV) {
-								opline->opcode = ZEND_CHECK_VAR;
-								SET_UNUSED(opline->result);
-							} else if (opline->op1_type & (IS_TMP_VAR|IS_VAR)) {
-								opline->opcode = ZEND_FREE;
-								SET_UNUSED(opline->result);
-							} else {
-								if (opline->op1_type == IS_CONST) {
-									literal_dtor(&ZEND_OP1_LITERAL(opline));
-								}
-								MAKE_NOP(opline);
-							}
+							zend_optimizer_convert_to_free_op1(op_array, opline);
 							break;
 						case ZEND_JMPZ_EX:
 						case ZEND_JMPNZ_EX:
@@ -1868,10 +1670,7 @@ void zend_optimize_cfg(zend_op_array *op_array, zend_optimizer_ctx *ctx)
 
     /* Build CFG */
 	checkpoint = zend_arena_checkpoint(ctx->arena);
-	if (zend_build_cfg(&ctx->arena, op_array, 0, &cfg) != SUCCESS) {
-		zend_arena_release(&ctx->arena, checkpoint);
-		return;
-	}
+	zend_build_cfg(&ctx->arena, op_array, 0, &cfg);
 
 	if (cfg.blocks_count * (op_array->last_var + op_array->T) > 64 * 1024 * 1024) {
 		zend_arena_release(&ctx->arena, checkpoint);
