@@ -5,7 +5,7 @@
    | This source file is subject to version 3.01 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
    | available through the world-wide-web at the following url:           |
-   | http://www.php.net/license/3_01.txt                                  |
+   | https://www.php.net/license/3_01.txt                                 |
    | If you did not receive a copy of the PHP license and are unable to   |
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
@@ -23,10 +23,8 @@
 #include "php.h"
 
 #include "zend_variables.h"
-#include "ext/standard/php_string.h"
 #include "ext/standard/info.h"
 #include "ext/standard/html.h"
-#include "zend_interfaces.h"
 
 #ifdef HAVE_XML
 
@@ -123,14 +121,6 @@ typedef struct {
 	unsigned short (*encoding_function)(unsigned char);
 } xml_encoding;
 
-
-enum php_xml_option {
-    PHP_XML_OPTION_CASE_FOLDING = 1,
-    PHP_XML_OPTION_TARGET_ENCODING,
-    PHP_XML_OPTION_SKIP_TAGSTART,
-    PHP_XML_OPTION_SKIP_WHITE
-};
-
 /* {{{ dynamically loadable module stuff */
 #ifdef COMPILE_DL_XML
 #ifdef ZTS
@@ -166,8 +156,8 @@ inline static char xml_decode_us_ascii(unsigned short);
 static void xml_call_handler(xml_parser *, zval *, zend_function *, int, zval *, zval *);
 static void _xml_xmlchar_zval(const XML_Char *, int, const XML_Char *, zval *);
 static int _xml_xmlcharlen(const XML_Char *);
-static void _xml_add_to_info(xml_parser *parser,char *name);
-inline static zend_string *_xml_decode_tag(xml_parser *parser, const char *tag);
+static void _xml_add_to_info(xml_parser *parser, const char *name);
+inline static zend_string *_xml_decode_tag(xml_parser *parser, const XML_Char *tag);
 
 void _xml_startElementHandler(void *, const XML_Char *, const XML_Char **);
 void _xml_endElementHandler(void *, const XML_Char *);
@@ -191,10 +181,10 @@ static const zend_module_dep xml_deps[] = {
 
 zend_module_entry xml_module_entry = {
 #ifdef LIBXML_EXPAT_COMPAT
-    STANDARD_MODULE_HEADER_EX, NULL,
+	STANDARD_MODULE_HEADER_EX, NULL,
 	xml_deps,
 #else
-    STANDARD_MODULE_HEADER,
+	STANDARD_MODULE_HEADER,
 #endif
 	"xml",                /* extension name */
 	ext_functions,        /* extension function list */
@@ -203,11 +193,11 @@ zend_module_entry xml_module_entry = {
 	NULL,                 /* per-request startup function */
 	NULL,                 /* per-request shutdown function */
 	PHP_MINFO(xml),       /* information function */
-    PHP_XML_VERSION,
-    PHP_MODULE_GLOBALS(xml), /* globals descriptor */
-    PHP_GINIT(xml),          /* globals ctor */
-    NULL,                    /* globals dtor */
-    NULL,                    /* post deactivate */
+	PHP_XML_VERSION,
+	PHP_MODULE_GLOBALS(xml), /* globals descriptor */
+	PHP_GINIT(xml),          /* globals ctor */
+	NULL,                    /* globals dtor */
+	NULL,                    /* post deactivate */
 	STANDARD_MODULE_PROPERTIES_EX
 };
 
@@ -253,13 +243,8 @@ static void php_xml_free_wrapper(void *ptr)
 
 PHP_MINIT_FUNCTION(xml)
 {
-	zend_class_entry ce;
-	INIT_CLASS_ENTRY(ce, "XMLParser", class_XMLParser_methods);
-	xml_parser_ce = zend_register_internal_class(&ce);
+	xml_parser_ce = register_class_XMLParser();
 	xml_parser_ce->create_object = xml_parser_create_object;
-	xml_parser_ce->ce_flags |= ZEND_ACC_FINAL | ZEND_ACC_NO_DYNAMIC_PROPERTIES;
-	xml_parser_ce->serialize = zend_class_serialize_deny;
-	xml_parser_ce->unserialize = zend_class_unserialize_deny;
 
 	memcpy(&xml_parser_object_handlers, &std_object_handlers, sizeof(zend_object_handlers));
 	xml_parser_object_handlers.offset = XtOffsetOf(xml_parser, std);
@@ -267,34 +252,9 @@ PHP_MINIT_FUNCTION(xml)
 	xml_parser_object_handlers.get_gc = xml_parser_get_gc;
 	xml_parser_object_handlers.get_constructor = xml_parser_get_constructor;
 	xml_parser_object_handlers.clone_obj = NULL;
+	xml_parser_object_handlers.compare = zend_objects_not_comparable;
 
-	REGISTER_LONG_CONSTANT("XML_ERROR_NONE", XML_ERROR_NONE, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_NO_MEMORY", XML_ERROR_NO_MEMORY, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_SYNTAX", XML_ERROR_SYNTAX, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_NO_ELEMENTS", XML_ERROR_NO_ELEMENTS, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_INVALID_TOKEN", XML_ERROR_INVALID_TOKEN, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_UNCLOSED_TOKEN", XML_ERROR_UNCLOSED_TOKEN, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_PARTIAL_CHAR", XML_ERROR_PARTIAL_CHAR, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_TAG_MISMATCH", XML_ERROR_TAG_MISMATCH, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_DUPLICATE_ATTRIBUTE", XML_ERROR_DUPLICATE_ATTRIBUTE, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_JUNK_AFTER_DOC_ELEMENT", XML_ERROR_JUNK_AFTER_DOC_ELEMENT, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_PARAM_ENTITY_REF", XML_ERROR_PARAM_ENTITY_REF, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_UNDEFINED_ENTITY", XML_ERROR_UNDEFINED_ENTITY, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_RECURSIVE_ENTITY_REF", XML_ERROR_RECURSIVE_ENTITY_REF, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_ASYNC_ENTITY", XML_ERROR_ASYNC_ENTITY, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_BAD_CHAR_REF", XML_ERROR_BAD_CHAR_REF, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_BINARY_ENTITY_REF", XML_ERROR_BINARY_ENTITY_REF, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_ATTRIBUTE_EXTERNAL_ENTITY_REF", XML_ERROR_ATTRIBUTE_EXTERNAL_ENTITY_REF, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_MISPLACED_XML_PI", XML_ERROR_MISPLACED_XML_PI, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_UNKNOWN_ENCODING", XML_ERROR_UNKNOWN_ENCODING, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_INCORRECT_ENCODING", XML_ERROR_INCORRECT_ENCODING, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_UNCLOSED_CDATA_SECTION", XML_ERROR_UNCLOSED_CDATA_SECTION, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_ERROR_EXTERNAL_ENTITY_HANDLING", XML_ERROR_EXTERNAL_ENTITY_HANDLING, CONST_CS|CONST_PERSISTENT);
-
-	REGISTER_LONG_CONSTANT("XML_OPTION_CASE_FOLDING", PHP_XML_OPTION_CASE_FOLDING, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_OPTION_TARGET_ENCODING", PHP_XML_OPTION_TARGET_ENCODING, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_OPTION_SKIP_TAGSTART", PHP_XML_OPTION_SKIP_TAGSTART, CONST_CS|CONST_PERSISTENT);
-	REGISTER_LONG_CONSTANT("XML_OPTION_SKIP_WHITE", PHP_XML_OPTION_SKIP_WHITE, CONST_CS|CONST_PERSISTENT);
+	register_xml_symbols(module_number);
 
 	/* this object should not be pre-initialised at compile time,
 	   as the order of members may vary */
@@ -437,7 +397,7 @@ static void xml_set_handler(zval *handler, zval *data)
 
 	/* IS_ARRAY might indicate that we're using array($obj, 'method') syntax */
 	if (Z_TYPE_P(data) != IS_ARRAY && Z_TYPE_P(data) != IS_OBJECT) {
-		convert_to_string_ex(data);
+		convert_to_string(data);
 		if (Z_STRLEN_P(data) == 0) {
 			ZVAL_UNDEF(handler);
 			return;
@@ -556,8 +516,8 @@ static zend_string *xml_utf8_decode(const XML_Char *s, size_t len, const XML_Cha
 	str = zend_string_alloc(len, 0);
 	ZSTR_LEN(str) = 0;
 	while (pos < len) {
-		int status = FAILURE;
-		c = php_next_utf8_char((const unsigned char*)s, (size_t) len, &pos, &status);
+		zend_result status = FAILURE;
+		c = php_next_utf8_char((const unsigned char*)s, len, &pos, &status);
 
 		if (status == FAILURE || c > 0xFFU) {
 			c = '?';
@@ -588,7 +548,7 @@ static int _xml_xmlcharlen(const XML_Char *s)
 /* }}} */
 
 /* {{{ _xml_add_to_info() */
-static void _xml_add_to_info(xml_parser *parser,char *name)
+static void _xml_add_to_info(xml_parser *parser, const char *name)
 {
 	zval *element;
 
@@ -596,10 +556,11 @@ static void _xml_add_to_info(xml_parser *parser,char *name)
 		return;
 	}
 
-	if ((element = zend_hash_str_find(Z_ARRVAL(parser->info), name, strlen(name))) == NULL) {
+	size_t name_len = strlen(name);
+	if ((element = zend_hash_str_find(Z_ARRVAL(parser->info), name, name_len)) == NULL) {
 		zval values;
 		array_init(&values);
-		element = zend_hash_str_update(Z_ARRVAL(parser->info), name, strlen(name), &values);
+		element = zend_hash_str_update(Z_ARRVAL(parser->info), name, name_len, &values);
 	}
 
 	add_next_index_long(element, parser->curtag);
@@ -609,14 +570,14 @@ static void _xml_add_to_info(xml_parser *parser,char *name)
 /* }}} */
 
 /* {{{ _xml_decode_tag() */
-static zend_string *_xml_decode_tag(xml_parser *parser, const char *tag)
+static zend_string *_xml_decode_tag(xml_parser *parser, const XML_Char *tag)
 {
 	zend_string *str;
 
-	str = xml_utf8_decode((const XML_Char *)tag, strlen(tag), parser->target_encoding);
+	str = xml_utf8_decode(tag, _xml_xmlcharlen(tag), parser->target_encoding);
 
 	if (parser->case_folding) {
-		php_strtoupper(ZSTR_VAL(str), ZSTR_LEN(str));
+		zend_str_toupper(ZSTR_VAL(str), ZSTR_LEN(str));
 	}
 
 	return str;
@@ -631,82 +592,84 @@ void _xml_startElementHandler(void *userData, const XML_Char *name, const XML_Ch
 	zend_string *att, *tag_name, *val;
 	zval retval, args[3];
 
-	if (parser) {
-		parser->level++;
+	if (!parser) {
+		return;
+	}
 
-		tag_name = _xml_decode_tag(parser, (const char *)name);
+	parser->level++;
 
-		if (!Z_ISUNDEF(parser->startElementHandler)) {
-			ZVAL_COPY(&args[0], &parser->index);
-			ZVAL_STRING(&args[1], SKIP_TAGSTART(ZSTR_VAL(tag_name)));
-			array_init(&args[2]);
+	tag_name = _xml_decode_tag(parser, name);
+
+	if (!Z_ISUNDEF(parser->startElementHandler)) {
+		ZVAL_COPY(&args[0], &parser->index);
+		ZVAL_STRING(&args[1], SKIP_TAGSTART(ZSTR_VAL(tag_name)));
+		array_init(&args[2]);
+
+		while (attributes && *attributes) {
+			zval tmp;
+
+			att = _xml_decode_tag(parser, attributes[0]);
+			val = xml_utf8_decode(attributes[1], strlen((char *)attributes[1]), parser->target_encoding);
+
+			ZVAL_STR(&tmp, val);
+			zend_symtable_update(Z_ARRVAL(args[2]), att, &tmp);
+
+			attributes += 2;
+
+			zend_string_release_ex(att, 0);
+		}
+
+		xml_call_handler(parser, &parser->startElementHandler, parser->startElementPtr, 3, args, &retval);
+		zval_ptr_dtor(&retval);
+	}
+
+	if (!Z_ISUNDEF(parser->data)) {
+		if (parser->level <= XML_MAXLEVEL)  {
+			zval tag, atr;
+			int atcnt = 0;
+
+			array_init(&tag);
+			array_init(&atr);
+
+			_xml_add_to_info(parser, ZSTR_VAL(tag_name) + parser->toffset);
+
+			add_assoc_string(&tag, "tag", SKIP_TAGSTART(ZSTR_VAL(tag_name))); /* cast to avoid gcc-warning */
+			add_assoc_string(&tag, "type", "open");
+			add_assoc_long(&tag, "level", parser->level);
+
+			parser->ltags[parser->level-1] = estrdup(ZSTR_VAL(tag_name));
+			parser->lastwasopen = 1;
+
+			attributes = (const XML_Char **) attrs;
 
 			while (attributes && *attributes) {
 				zval tmp;
 
-				att = _xml_decode_tag(parser, (const char *)attributes[0]);
+				att = _xml_decode_tag(parser, attributes[0]);
 				val = xml_utf8_decode(attributes[1], strlen((char *)attributes[1]), parser->target_encoding);
 
 				ZVAL_STR(&tmp, val);
-				zend_symtable_update(Z_ARRVAL(args[2]), att, &tmp);
+				zend_symtable_update(Z_ARRVAL(atr), att, &tmp);
 
+				atcnt++;
 				attributes += 2;
 
 				zend_string_release_ex(att, 0);
 			}
 
-			xml_call_handler(parser, &parser->startElementHandler, parser->startElementPtr, 3, args, &retval);
-			zval_ptr_dtor(&retval);
-		}
-
-		if (!Z_ISUNDEF(parser->data)) {
-			if (parser->level <= XML_MAXLEVEL)  {
-				zval tag, atr;
-				int atcnt = 0;
-
-				array_init(&tag);
-				array_init(&atr);
-
-				_xml_add_to_info(parser, ZSTR_VAL(tag_name) + parser->toffset);
-
-				add_assoc_string(&tag, "tag", SKIP_TAGSTART(ZSTR_VAL(tag_name))); /* cast to avoid gcc-warning */
-				add_assoc_string(&tag, "type", "open");
-				add_assoc_long(&tag, "level", parser->level);
-
-				parser->ltags[parser->level-1] = estrdup(ZSTR_VAL(tag_name));
-				parser->lastwasopen = 1;
-
-				attributes = (const XML_Char **) attrs;
-
-				while (attributes && *attributes) {
-					zval tmp;
-
-					att = _xml_decode_tag(parser, (const char *)attributes[0]);
-					val = xml_utf8_decode(attributes[1], strlen((char *)attributes[1]), parser->target_encoding);
-
-					ZVAL_STR(&tmp, val);
-					zend_symtable_update(Z_ARRVAL(atr), att, &tmp);
-
-					atcnt++;
-					attributes += 2;
-
-					zend_string_release_ex(att, 0);
-				}
-
-				if (atcnt) {
-					zend_hash_str_add(Z_ARRVAL(tag), "attributes", sizeof("attributes") - 1, &atr);
-				} else {
-					zval_ptr_dtor(&atr);
-				}
-
-				parser->ctag = zend_hash_next_index_insert(Z_ARRVAL(parser->data), &tag);
-			} else if (parser->level == (XML_MAXLEVEL + 1)) {
-							php_error_docref(NULL, E_WARNING, "Maximum depth exceeded - Results truncated");
+			if (atcnt) {
+				zend_hash_str_add(Z_ARRVAL(tag), "attributes", sizeof("attributes") - 1, &atr);
+			} else {
+				zval_ptr_dtor(&atr);
 			}
-		}
 
-		zend_string_release_ex(tag_name, 0);
+			parser->ctag = zend_hash_next_index_insert(Z_ARRVAL(parser->data), &tag);
+		} else if (parser->level == (XML_MAXLEVEL + 1)) {
+						php_error_docref(NULL, E_WARNING, "Maximum depth exceeded - Results truncated");
+		}
 	}
+
+	zend_string_release_ex(tag_name, 0);
 }
 /* }}} */
 
@@ -715,47 +678,49 @@ void _xml_endElementHandler(void *userData, const XML_Char *name)
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser) {
-		zval retval, args[2];
-
-		zend_string *tag_name = _xml_decode_tag(parser, (const char *)name);
-
-		if (!Z_ISUNDEF(parser->endElementHandler)) {
-			ZVAL_COPY(&args[0], &parser->index);
-			ZVAL_STRING(&args[1], SKIP_TAGSTART(ZSTR_VAL(tag_name)));
-
-			xml_call_handler(parser, &parser->endElementHandler, parser->endElementPtr, 2, args, &retval);
-			zval_ptr_dtor(&retval);
-		}
-
-		if (!Z_ISUNDEF(parser->data)) {
-			zval tag;
-
-			if (parser->lastwasopen) {
-				add_assoc_string(parser->ctag, "type", "complete");
-			} else {
-				array_init(&tag);
-
-				_xml_add_to_info(parser, ZSTR_VAL(tag_name) + parser->toffset);
-
-				add_assoc_string(&tag, "tag", SKIP_TAGSTART(ZSTR_VAL(tag_name))); /* cast to avoid gcc-warning */
-				add_assoc_string(&tag, "type", "close");
-				add_assoc_long(&tag, "level", parser->level);
-
-				zend_hash_next_index_insert(Z_ARRVAL(parser->data), &tag);
-			}
-
-			parser->lastwasopen = 0;
-		}
-
-		zend_string_release_ex(tag_name, 0);
-
-		if ((parser->ltags) && (parser->level <= XML_MAXLEVEL)) {
-			efree(parser->ltags[parser->level-1]);
-		}
-
-		parser->level--;
+	if (!parser) {
+		return;
 	}
+
+	zval retval, args[2];
+
+	zend_string *tag_name = _xml_decode_tag(parser, name);
+
+	if (!Z_ISUNDEF(parser->endElementHandler)) {
+		ZVAL_COPY(&args[0], &parser->index);
+		ZVAL_STRING(&args[1], SKIP_TAGSTART(ZSTR_VAL(tag_name)));
+
+		xml_call_handler(parser, &parser->endElementHandler, parser->endElementPtr, 2, args, &retval);
+		zval_ptr_dtor(&retval);
+	}
+
+	if (!Z_ISUNDEF(parser->data)) {
+		zval tag;
+
+		if (parser->lastwasopen) {
+			add_assoc_string(parser->ctag, "type", "complete");
+		} else {
+			array_init(&tag);
+
+			_xml_add_to_info(parser, ZSTR_VAL(tag_name) + parser->toffset);
+
+			add_assoc_string(&tag, "tag", SKIP_TAGSTART(ZSTR_VAL(tag_name))); /* cast to avoid gcc-warning */
+			add_assoc_string(&tag, "type", "close");
+			add_assoc_long(&tag, "level", parser->level);
+
+			zend_hash_next_index_insert(Z_ARRVAL(parser->data), &tag);
+		}
+
+		parser->lastwasopen = 0;
+	}
+
+	zend_string_release_ex(tag_name, 0);
+
+	if ((parser->ltags) && (parser->level <= XML_MAXLEVEL)) {
+		efree(parser->ltags[parser->level-1]);
+	}
+
+	parser->level--;
 }
 /* }}} */
 
@@ -764,89 +729,88 @@ void _xml_characterDataHandler(void *userData, const XML_Char *s, int len)
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser) {
-		zval retval, args[2];
+	if (!parser) {
+		return;
+	}
 
-		if (!Z_ISUNDEF(parser->characterDataHandler)) {
-			ZVAL_COPY(&args[0], &parser->index);
-			_xml_xmlchar_zval(s, len, parser->target_encoding, &args[1]);
-			xml_call_handler(parser, &parser->characterDataHandler, parser->characterDataPtr, 2, args, &retval);
-			zval_ptr_dtor(&retval);
-		}
+	zval retval, args[2];
 
-		if (!Z_ISUNDEF(parser->data)) {
-			size_t i;
-			int doprint = 0;
-			zend_string *decoded_value;
+	if (!Z_ISUNDEF(parser->characterDataHandler)) {
+		ZVAL_COPY(&args[0], &parser->index);
+		_xml_xmlchar_zval(s, len, parser->target_encoding, &args[1]);
+		xml_call_handler(parser, &parser->characterDataHandler, parser->characterDataPtr, 2, args, &retval);
+		zval_ptr_dtor(&retval);
+	}
 
-			decoded_value = xml_utf8_decode(s, len, parser->target_encoding);
-			for (i = 0; i < ZSTR_LEN(decoded_value); i++) {
-				switch (ZSTR_VAL(decoded_value)[i]) {
-					case ' ':
-					case '\t':
-					case '\n':
-						continue;
-					default:
-						doprint = 1;
-						break;
-				}
-				if (doprint) {
+	if (Z_ISUNDEF(parser->data)) {
+		return;
+	}
+
+	bool doprint = 0;
+	zend_string *decoded_value;
+	decoded_value = xml_utf8_decode(s, len, parser->target_encoding);
+	if (parser->skipwhite) {
+		for (size_t i = 0; i < ZSTR_LEN(decoded_value); i++) {
+			switch (ZSTR_VAL(decoded_value)[i]) {
+				case ' ':
+				case '\t':
+				case '\n':
+					continue;
+				default:
+					doprint = 1;
 					break;
-				}
 			}
+			if (doprint) {
+				break;
+			}
+		}
+	}
+	if (parser->lastwasopen) {
+		zval *myval;
+		/* check if the current tag already has a value - if yes append to that! */
+		if ((myval = zend_hash_str_find(Z_ARRVAL_P(parser->ctag), "value", sizeof("value") - 1))) {
+			size_t newlen = Z_STRLEN_P(myval) + ZSTR_LEN(decoded_value);
+			Z_STR_P(myval) = zend_string_extend(Z_STR_P(myval), newlen, 0);
+			strncpy(Z_STRVAL_P(myval) + Z_STRLEN_P(myval) - ZSTR_LEN(decoded_value),
+					ZSTR_VAL(decoded_value), ZSTR_LEN(decoded_value) + 1);
+			zend_string_release_ex(decoded_value, 0);
+		} else {
 			if (doprint || (! parser->skipwhite)) {
-				if (parser->lastwasopen) {
-					zval *myval;
-
-					/* check if the current tag already has a value - if yes append to that! */
-					if ((myval = zend_hash_str_find(Z_ARRVAL_P(parser->ctag), "value", sizeof("value") - 1))) {
-						int newlen = Z_STRLEN_P(myval) + ZSTR_LEN(decoded_value);
+				add_assoc_str(parser->ctag, "value", decoded_value);
+			} else {
+				zend_string_release_ex(decoded_value, 0);
+			}
+		}
+	} else {
+		zval tag;
+		zval *curtag, *mytype, *myval;
+		ZEND_HASH_REVERSE_FOREACH_VAL(Z_ARRVAL(parser->data), curtag) {
+			if ((mytype = zend_hash_str_find(Z_ARRVAL_P(curtag),"type", sizeof("type") - 1))) {
+				if (zend_string_equals_literal(Z_STR_P(mytype), "cdata")) {
+					if ((myval = zend_hash_str_find(Z_ARRVAL_P(curtag), "value", sizeof("value") - 1))) {
+						size_t newlen = Z_STRLEN_P(myval) + ZSTR_LEN(decoded_value);
 						Z_STR_P(myval) = zend_string_extend(Z_STR_P(myval), newlen, 0);
 						strncpy(Z_STRVAL_P(myval) + Z_STRLEN_P(myval) - ZSTR_LEN(decoded_value),
 								ZSTR_VAL(decoded_value), ZSTR_LEN(decoded_value) + 1);
 						zend_string_release_ex(decoded_value, 0);
-					} else {
-						add_assoc_str(parser->ctag, "value", decoded_value);
-					}
-
-				} else {
-					zval tag;
-					zval *curtag, *mytype, *myval;
-
-					ZEND_HASH_REVERSE_FOREACH_VAL(Z_ARRVAL(parser->data), curtag) {
-						if ((mytype = zend_hash_str_find(Z_ARRVAL_P(curtag),"type", sizeof("type") - 1))) {
-							if (!strcmp(Z_STRVAL_P(mytype), "cdata")) {
-								if ((myval = zend_hash_str_find(Z_ARRVAL_P(curtag), "value", sizeof("value") - 1))) {
-									int newlen = Z_STRLEN_P(myval) + ZSTR_LEN(decoded_value);
-									Z_STR_P(myval) = zend_string_extend(Z_STR_P(myval), newlen, 0);
-									strncpy(Z_STRVAL_P(myval) + Z_STRLEN_P(myval) - ZSTR_LEN(decoded_value),
-											ZSTR_VAL(decoded_value), ZSTR_LEN(decoded_value) + 1);
-									zend_string_release_ex(decoded_value, 0);
-									return;
-								}
-							}
-						}
-						break;
-					} ZEND_HASH_FOREACH_END();
-
-					if (parser->level <= XML_MAXLEVEL && parser->level > 0) {
-						array_init(&tag);
-
-						_xml_add_to_info(parser,SKIP_TAGSTART(parser->ltags[parser->level-1]));
-
-						add_assoc_string(&tag, "tag", SKIP_TAGSTART(parser->ltags[parser->level-1]));
-						add_assoc_str(&tag, "value", decoded_value);
-						add_assoc_string(&tag, "type", "cdata");
-						add_assoc_long(&tag, "level", parser->level);
-
-						zend_hash_next_index_insert(Z_ARRVAL(parser->data), &tag);
-					} else if (parser->level == (XML_MAXLEVEL + 1)) {
-											php_error_docref(NULL, E_WARNING, "Maximum depth exceeded - Results truncated");
+						return;
 					}
 				}
-			} else {
-				zend_string_release_ex(decoded_value, 0);
 			}
+			break;
+		} ZEND_HASH_FOREACH_END();
+		if (parser->level <= XML_MAXLEVEL && parser->level > 0 && (doprint || (! parser->skipwhite))) {
+			array_init(&tag);
+			_xml_add_to_info(parser,SKIP_TAGSTART(parser->ltags[parser->level-1]));
+			add_assoc_string(&tag, "tag", SKIP_TAGSTART(parser->ltags[parser->level-1]));
+			add_assoc_str(&tag, "value", decoded_value);
+			add_assoc_string(&tag, "type", "cdata");
+			add_assoc_long(&tag, "level", parser->level);
+			zend_hash_next_index_insert(Z_ARRVAL(parser->data), &tag);
+		} else if (parser->level == (XML_MAXLEVEL + 1)) {
+								php_error_docref(NULL, E_WARNING, "Maximum depth exceeded - Results truncated");
+		} else {
+			zend_string_release_ex(decoded_value, 0);
 		}
 	}
 }
@@ -857,15 +821,17 @@ void _xml_processingInstructionHandler(void *userData, const XML_Char *target, c
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser && !Z_ISUNDEF(parser->processingInstructionHandler)) {
-		zval retval, args[3];
-
-		ZVAL_COPY(&args[0], &parser->index);
-		_xml_xmlchar_zval(target, 0, parser->target_encoding, &args[1]);
-		_xml_xmlchar_zval(data, 0, parser->target_encoding, &args[2]);
-		xml_call_handler(parser, &parser->processingInstructionHandler, parser->processingInstructionPtr, 3, args, &retval);
-		zval_ptr_dtor(&retval);
+	if (!parser || Z_ISUNDEF(parser->processingInstructionHandler)) {
+		return;
 	}
+
+	zval retval, args[3];
+
+	ZVAL_COPY(&args[0], &parser->index);
+	_xml_xmlchar_zval(target, 0, parser->target_encoding, &args[1]);
+	_xml_xmlchar_zval(data, 0, parser->target_encoding, &args[2]);
+	xml_call_handler(parser, &parser->processingInstructionHandler, parser->processingInstructionPtr, 3, args, &retval);
+	zval_ptr_dtor(&retval);
 }
 /* }}} */
 
@@ -874,91 +840,91 @@ void _xml_defaultHandler(void *userData, const XML_Char *s, int len)
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser && !Z_ISUNDEF(parser->defaultHandler)) {
-		zval retval, args[2];
-
-		ZVAL_COPY(&args[0], &parser->index);
-		_xml_xmlchar_zval(s, len, parser->target_encoding, &args[1]);
-		xml_call_handler(parser, &parser->defaultHandler, parser->defaultPtr, 2, args, &retval);
-		zval_ptr_dtor(&retval);
+	if (!parser || Z_ISUNDEF(parser->defaultHandler)) {
+		return;
 	}
+
+	zval retval, args[2];
+
+	ZVAL_COPY(&args[0], &parser->index);
+	_xml_xmlchar_zval(s, len, parser->target_encoding, &args[1]);
+	xml_call_handler(parser, &parser->defaultHandler, parser->defaultPtr, 2, args, &retval);
+	zval_ptr_dtor(&retval);
 }
 /* }}} */
 
 /* {{{ _xml_unparsedEntityDeclHandler() */
 void _xml_unparsedEntityDeclHandler(void *userData,
-										 const XML_Char *entityName,
-										 const XML_Char *base,
-										 const XML_Char *systemId,
-										 const XML_Char *publicId,
-										 const XML_Char *notationName)
+	const XML_Char *entityName, const XML_Char *base, const XML_Char *systemId,
+	const XML_Char *publicId, const XML_Char *notationName)
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser && !Z_ISUNDEF(parser->unparsedEntityDeclHandler)) {
-		zval retval, args[6];
-
-		ZVAL_COPY(&args[0], &parser->index);
-		_xml_xmlchar_zval(entityName, 0, parser->target_encoding, &args[1]);
-		_xml_xmlchar_zval(base, 0, parser->target_encoding, &args[2]);
-		_xml_xmlchar_zval(systemId, 0, parser->target_encoding, &args[3]);
-		_xml_xmlchar_zval(publicId, 0, parser->target_encoding, &args[4]);
-		_xml_xmlchar_zval(notationName, 0, parser->target_encoding, &args[5]);
-		xml_call_handler(parser, &parser->unparsedEntityDeclHandler, parser->unparsedEntityDeclPtr, 6, args, &retval);
-		zval_ptr_dtor(&retval);
+	if (!parser || Z_ISUNDEF(parser->unparsedEntityDeclHandler)) {
+		return;
 	}
+
+	zval retval, args[6];
+
+	ZVAL_COPY(&args[0], &parser->index);
+	_xml_xmlchar_zval(entityName, 0, parser->target_encoding, &args[1]);
+	_xml_xmlchar_zval(base, 0, parser->target_encoding, &args[2]);
+	_xml_xmlchar_zval(systemId, 0, parser->target_encoding, &args[3]);
+	_xml_xmlchar_zval(publicId, 0, parser->target_encoding, &args[4]);
+	_xml_xmlchar_zval(notationName, 0, parser->target_encoding, &args[5]);
+	xml_call_handler(parser, &parser->unparsedEntityDeclHandler, parser->unparsedEntityDeclPtr, 6, args, &retval);
+	zval_ptr_dtor(&retval);
 }
 /* }}} */
 
 /* {{{ _xml_notationDeclHandler() */
-void _xml_notationDeclHandler(void *userData,
-							  const XML_Char *notationName,
-							  const XML_Char *base,
-							  const XML_Char *systemId,
-							  const XML_Char *publicId)
+void _xml_notationDeclHandler(void *userData, const XML_Char *notationName,
+	const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId)
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser && !Z_ISUNDEF(parser->notationDeclHandler)) {
-		zval retval, args[5];
-
-		ZVAL_COPY(&args[0], &parser->index);
-		_xml_xmlchar_zval(notationName, 0, parser->target_encoding, &args[1]);
-		_xml_xmlchar_zval(base, 0, parser->target_encoding, &args[2]);
-		_xml_xmlchar_zval(systemId, 0, parser->target_encoding, &args[3]);
-		_xml_xmlchar_zval(publicId, 0, parser->target_encoding, &args[4]);
-		xml_call_handler(parser, &parser->notationDeclHandler, parser->notationDeclPtr, 5, args, &retval);
-		zval_ptr_dtor(&retval);
+	if (!parser || Z_ISUNDEF(parser->notationDeclHandler)) {
+		return;
 	}
+
+	zval retval, args[5];
+
+	ZVAL_COPY(&args[0], &parser->index);
+	_xml_xmlchar_zval(notationName, 0, parser->target_encoding, &args[1]);
+	_xml_xmlchar_zval(base, 0, parser->target_encoding, &args[2]);
+	_xml_xmlchar_zval(systemId, 0, parser->target_encoding, &args[3]);
+	_xml_xmlchar_zval(publicId, 0, parser->target_encoding, &args[4]);
+	xml_call_handler(parser, &parser->notationDeclHandler, parser->notationDeclPtr, 5, args, &retval);
+	zval_ptr_dtor(&retval);
 }
 /* }}} */
 
 /* {{{ _xml_externalEntityRefHandler() */
-int _xml_externalEntityRefHandler(XML_Parser parserPtr,
-								   const XML_Char *openEntityNames,
-								   const XML_Char *base,
-								   const XML_Char *systemId,
-								   const XML_Char *publicId)
+int _xml_externalEntityRefHandler(XML_Parser parserPtr, const XML_Char *openEntityNames,
+	const XML_Char *base, const XML_Char *systemId, const XML_Char *publicId)
 {
 	xml_parser *parser = XML_GetUserData(parserPtr);
-	int ret = 0; /* abort if no handler is set (should be configurable?) */
 
-	if (parser && !Z_ISUNDEF(parser->externalEntityRefHandler)) {
-		zval retval, args[5];
-
-		ZVAL_COPY(&args[0], &parser->index);
-		_xml_xmlchar_zval(openEntityNames, 0, parser->target_encoding, &args[1]);
-		_xml_xmlchar_zval(base, 0, parser->target_encoding, &args[2]);
-		_xml_xmlchar_zval(systemId, 0, parser->target_encoding, &args[3]);
-		_xml_xmlchar_zval(publicId, 0, parser->target_encoding, &args[4]);
-		xml_call_handler(parser, &parser->externalEntityRefHandler, parser->externalEntityRefPtr, 5, args, &retval);
-		if (!Z_ISUNDEF(retval)) {
-			convert_to_long(&retval);
-			ret = Z_LVAL(retval);
-		} else {
-			ret = 0;
-		}
+	if (!parser || Z_ISUNDEF(parser->externalEntityRefHandler)) {
+		return 0;
 	}
+
+	int ret = 0; /* abort if no handler is set (should be configurable?) */
+	zval retval, args[5];
+
+	ZVAL_COPY(&args[0], &parser->index);
+	_xml_xmlchar_zval(openEntityNames, 0, parser->target_encoding, &args[1]);
+	_xml_xmlchar_zval(base, 0, parser->target_encoding, &args[2]);
+	_xml_xmlchar_zval(systemId, 0, parser->target_encoding, &args[3]);
+	_xml_xmlchar_zval(publicId, 0, parser->target_encoding, &args[4]);
+	xml_call_handler(parser, &parser->externalEntityRefHandler, parser->externalEntityRefPtr, 5, args, &retval);
+	if (!Z_ISUNDEF(retval)) {
+		convert_to_long(&retval);
+		ret = Z_LVAL(retval);
+	} else {
+		ret = 0;
+	}
+
 	return ret;
 }
 /* }}} */
@@ -968,15 +934,17 @@ void _xml_startNamespaceDeclHandler(void *userData,const XML_Char *prefix, const
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser && !Z_ISUNDEF(parser->startNamespaceDeclHandler)) {
-		zval retval, args[3];
-
-		ZVAL_COPY(&args[0], &parser->index);
-		_xml_xmlchar_zval(prefix, 0, parser->target_encoding, &args[1]);
-		_xml_xmlchar_zval(uri, 0, parser->target_encoding, &args[2]);
-		xml_call_handler(parser, &parser->startNamespaceDeclHandler, parser->startNamespaceDeclPtr, 3, args, &retval);
-		zval_ptr_dtor(&retval);
+	if (!parser || Z_ISUNDEF(parser->startNamespaceDeclHandler)) {
+		return;
 	}
+
+	zval retval, args[3];
+
+	ZVAL_COPY(&args[0], &parser->index);
+	_xml_xmlchar_zval(prefix, 0, parser->target_encoding, &args[1]);
+	_xml_xmlchar_zval(uri, 0, parser->target_encoding, &args[2]);
+	xml_call_handler(parser, &parser->startNamespaceDeclHandler, parser->startNamespaceDeclPtr, 3, args, &retval);
+	zval_ptr_dtor(&retval);
 }
 /* }}} */
 
@@ -985,14 +953,16 @@ void _xml_endNamespaceDeclHandler(void *userData, const XML_Char *prefix)
 {
 	xml_parser *parser = (xml_parser *)userData;
 
-	if (parser && !Z_ISUNDEF(parser->endNamespaceDeclHandler)) {
-		zval retval, args[2];
-
-		ZVAL_COPY(&args[0], &parser->index);
-		_xml_xmlchar_zval(prefix, 0, parser->target_encoding, &args[1]);
-		xml_call_handler(parser, &parser->endNamespaceDeclHandler, parser->endNamespaceDeclPtr, 2, args, &retval);
-		zval_ptr_dtor(&retval);
+	if (!parser || Z_ISUNDEF(parser->endNamespaceDeclHandler)) {
+		return;
 	}
+
+	zval retval, args[2];
+
+	ZVAL_COPY(&args[0], &parser->index);
+	_xml_xmlchar_zval(prefix, 0, parser->target_encoding, &args[1]);
+	xml_call_handler(parser, &parser->endNamespaceDeclHandler, parser->endNamespaceDeclPtr, 2, args, &retval);
+	zval_ptr_dtor(&retval);
 }
 /* }}} */
 
@@ -1003,15 +973,14 @@ static void php_xml_parser_create_impl(INTERNAL_FUNCTION_PARAMETERS, int ns_supp
 	xml_parser *parser;
 	int auto_detect = 0;
 
-	char *encoding_param = NULL;
-	size_t encoding_param_len = 0;
+	zend_string *encoding_param = NULL;
 
 	char *ns_param = NULL;
 	size_t ns_param_len = 0;
 
 	XML_Char *encoding;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), (ns_support ? "|s!s": "|s!"), &encoding_param, &encoding_param_len, &ns_param, &ns_param_len) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), (ns_support ? "|S!s": "|S!"), &encoding_param, &ns_param, &ns_param_len) == FAILURE) {
 		RETURN_THROWS();
 	}
 
@@ -1019,14 +988,14 @@ static void php_xml_parser_create_impl(INTERNAL_FUNCTION_PARAMETERS, int ns_supp
 		/* The supported encoding types are hardcoded here because
 		 * we are limited to the encodings supported by expat/xmltok.
 		 */
-		if (encoding_param_len == 0) {
+		if (ZSTR_LEN(encoding_param) == 0) {
 			encoding = XML(default_encoding);
 			auto_detect = 1;
-		} else if (strcasecmp(encoding_param, "ISO-8859-1") == 0) {
+		} else if (zend_string_equals_literal_ci(encoding_param, "ISO-8859-1")) {
 			encoding = (XML_Char*)"ISO-8859-1";
-		} else if (strcasecmp(encoding_param, "UTF-8") == 0) {
+		} else if (zend_string_equals_literal_ci(encoding_param, "UTF-8")) {
 			encoding = (XML_Char*)"UTF-8";
-		} else if (strcasecmp(encoding_param, "US-ASCII") == 0) {
+		} else if (zend_string_equals_literal_ci(encoding_param, "US-ASCII")) {
 			encoding = (XML_Char*)"US-ASCII";
 		} else {
 			zend_argument_value_error(1, "is not a supported source encoding");
@@ -1043,7 +1012,7 @@ static void php_xml_parser_create_impl(INTERNAL_FUNCTION_PARAMETERS, int ns_supp
 	object_init_ex(return_value, xml_parser_ce);
 	parser = Z_XMLPARSER_P(return_value);
 	parser->parser = XML_ParserCreate_MM((auto_detect ? NULL : encoding),
-                                         &php_xml_mem_hdlrs, (XML_Char*)ns_param);
+	                                     &php_xml_mem_hdlrs, (XML_Char*)ns_param);
 
 	parser->target_encoding = encoding;
 	parser->case_folding = 1;
@@ -1083,7 +1052,7 @@ PHP_FUNCTION(xml_set_object)
 	zval_ptr_dtor(&parser->object);
 	ZVAL_OBJ_COPY(&parser->object, Z_OBJ_P(mythis));
 
-	RETVAL_TRUE;
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1101,7 +1070,8 @@ PHP_FUNCTION(xml_set_element_handler)
 	xml_set_handler(&parser->startElementHandler, shdl);
 	xml_set_handler(&parser->endElementHandler, ehdl);
 	XML_SetElementHandler(parser->parser, _xml_startElementHandler, _xml_endElementHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1118,7 +1088,8 @@ PHP_FUNCTION(xml_set_character_data_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->characterDataHandler, hdl);
 	XML_SetCharacterDataHandler(parser->parser, _xml_characterDataHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1135,7 +1106,8 @@ PHP_FUNCTION(xml_set_processing_instruction_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->processingInstructionHandler, hdl);
 	XML_SetProcessingInstructionHandler(parser->parser, _xml_processingInstructionHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1152,7 +1124,8 @@ PHP_FUNCTION(xml_set_default_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->defaultHandler, hdl);
 	XML_SetDefaultHandler(parser->parser, _xml_defaultHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1169,7 +1142,8 @@ PHP_FUNCTION(xml_set_unparsed_entity_decl_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->unparsedEntityDeclHandler, hdl);
 	XML_SetUnparsedEntityDeclHandler(parser->parser, _xml_unparsedEntityDeclHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1186,7 +1160,8 @@ PHP_FUNCTION(xml_set_notation_decl_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->notationDeclHandler, hdl);
 	XML_SetNotationDeclHandler(parser->parser, _xml_notationDeclHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1203,7 +1178,8 @@ PHP_FUNCTION(xml_set_external_entity_ref_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->externalEntityRefHandler, hdl);
 	XML_SetExternalEntityRefHandler(parser->parser, (void *) _xml_externalEntityRefHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1220,7 +1196,8 @@ PHP_FUNCTION(xml_set_start_namespace_decl_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->startNamespaceDeclHandler, hdl);
 	XML_SetStartNamespaceDeclHandler(parser->parser, _xml_startNamespaceDeclHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1237,7 +1214,8 @@ PHP_FUNCTION(xml_set_end_namespace_decl_handler)
 	parser = Z_XMLPARSER_P(pind);
 	xml_set_handler(&parser->endNamespaceDeclHandler, hdl);
 	XML_SetEndNamespaceDeclHandler(parser->parser, _xml_endNamespaceDeclHandler);
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1249,13 +1227,17 @@ PHP_FUNCTION(xml_parse)
 	char *data;
 	size_t data_len;
 	int ret;
-	zend_bool isFinal = 0;
+	bool isFinal = 0;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "Os|b", &pind, xml_parser_ce, &data, &data_len, &isFinal) == FAILURE) {
 		RETURN_THROWS();
 	}
 
 	parser = Z_XMLPARSER_P(pind);
+	if (parser->isparsing) {
+		zend_throw_error(NULL, "Parser must not be called recursively");
+		RETURN_THROWS();
+	}
 	parser->isparsing = 1;
 	ret = XML_Parse(parser->parser, (XML_Char*)data, data_len, isFinal);
 	parser->isparsing = 0;
@@ -1303,6 +1285,10 @@ PHP_FUNCTION(xml_parse_into_struct)
 	XML_SetElementHandler(parser->parser, _xml_startElementHandler, _xml_endElementHandler);
 	XML_SetCharacterDataHandler(parser->parser, _xml_characterDataHandler);
 
+	if (parser->isparsing) {
+		php_error_docref(NULL, E_WARNING, "Parser must not be called recursively");
+		RETURN_FALSE;
+	}
 	parser->isparsing = 1;
 	ret = XML_Parse(parser->parser, (XML_Char*)data, data_len, 1);
 	parser->isparsing = 0;
@@ -1400,7 +1386,7 @@ PHP_FUNCTION(xml_parser_free)
 
 	parser = Z_XMLPARSER_P(pind);
 	if (parser->isparsing == 1) {
-		php_error_docref(NULL, E_WARNING, "Parser cannot be freed while it is parsing.");
+		php_error_docref(NULL, E_WARNING, "Parser cannot be freed while it is parsing");
 		RETURN_FALSE;
 	}
 
@@ -1429,6 +1415,8 @@ PHP_FUNCTION(xml_parser_set_option)
 			if (parser->toffset < 0) {
 				php_error_docref(NULL, E_WARNING, "tagstart ignored, because it is out of range");
 				parser->toffset = 0;
+				/* TODO Promote to ValueError in PHP 9.0 */
+				RETURN_FALSE;
 			}
 			break;
 		case PHP_XML_OPTION_SKIP_WHITE:
@@ -1450,11 +1438,12 @@ PHP_FUNCTION(xml_parser_set_option)
 			break;
 		}
 		default:
-			zend_argument_value_error(2, "must be a PHP_XML_OPTION_* constant");
+			zend_argument_value_error(2, "must be a XML_OPTION_* constant");
 			RETURN_THROWS();
 			break;
 	}
-	RETVAL_TRUE;
+
+	RETURN_TRUE;
 }
 /* }}} */
 
@@ -1484,7 +1473,7 @@ PHP_FUNCTION(xml_parser_get_option)
 			RETURN_STRING((char *)parser->target_encoding);
 			break;
 		default:
-			zend_argument_value_error(2, "must be a PHP_XML_OPTION_* constant");
+			zend_argument_value_error(2, "must be a XML_OPTION_* constant");
 			RETURN_THROWS();
 	}
 }
